import json
import os
import string
from collections import OrderedDict
from decimal import Decimal
from operator import (add,
                      and_,
                      or_,
                      sub,
                      xor)
from typing import (Any,
                    Tuple)

from hypothesis import strategies
from hypothesis.searchstrategy import SearchStrategy

from lz.functional import (identity,
                           to_constant)
from lz.hints import Map
from .literals import empty
from .literals.base import (integers,
                            json_serializable_objects,
                            lists,
                            numbers,
                            objects,
                            real_numbers,
                            sets,
                            sortable_domains,
                            strings,
                            tuples)
from .literals.factories import (to_iterables,
                                 to_lists,
                                 to_strings)

false_predicates = strategies.just(to_constant(False))
true_predicates = strategies.just(to_constant(True))
predicates = false_predicates | true_predicates
predicates_arguments = objects
starting_maps = [identity, float, str, json.dumps]
suitable_maps = {identity: [identity, float, str],
                 float: starting_maps,
                 str: [identity, float, str, json.loads],
                 json.dumps: [identity, float, str, json.loads],
                 json.loads: starting_maps}
suitable_maps = dict(zip(suitable_maps.keys(),
                         map(strategies.sampled_from, suitable_maps.values())))
to_one_of_suitable_maps = suitable_maps.__getitem__
maps = strategies.sampled_from(starting_maps)
maps_lists = to_lists(maps)
maps_arguments = (strategies.integers()
                  | strategies.floats(allow_nan=False,
                                      allow_infinity=False))
maps_lists_arguments = to_lists(maps_arguments)


@strategies.composite
def extend_suitable_maps(draw: Map[SearchStrategy, Any],
                         maps_tuples: SearchStrategy) -> Tuple[Map, ...]:
    maps_tuple = draw(maps_tuples)
    last_map = draw(to_one_of_suitable_maps(maps_tuple[0]))
    return (last_map,) + maps_tuple


suitable_maps = strategies.recursive(strategies.tuples(maps),
                                     extend_suitable_maps)
# "transparent" is an abbr. of "referential transparent"
transparent_functions = strategies.sampled_from([bool, complex, float,
                                                 identity, int,
                                                 json.dumps, json.loads,
                                                 os.path.join, str])
paths_names_parts = to_strings(string.digits + string.ascii_letters + '_')
transparent_functions_args = {
    bool: strategies.tuples(objects),
    complex: strategies.tuples(numbers),
    float: strategies.tuples(real_numbers),
    identity: strategies.tuples(objects),
    int: strategies.tuples(integers),
    json.dumps: strategies.tuples(json_serializable_objects),
    json.loads: strategies.tuples(json_serializable_objects.map(json.dumps)),
    os.path.join: to_lists(paths_names_parts,
                           min_size=1).map(tuple),
    str: strategies.tuples(objects),
    sum: strategies.tuples(to_iterables(numbers)),
}
transparent_functions_kwargs = {
    bool: empty.dictionaries,
    complex: empty.dictionaries,
    float: empty.dictionaries,
    identity: empty.dictionaries,
    int: empty.dictionaries,
    json.dumps: strategies.fixed_dictionaries(
            {'indent': strategies.integers(0, 10),
             'sort_keys': strategies.booleans()}),
    json.loads: strategies.fixed_dictionaries({
        'object_pairs_hook': strategies.none() | strategies.just(OrderedDict),
        'parse_float': strategies.none() | strategies.sampled_from([float,
                                                                    Decimal]),
        'parse_int': strategies.none() | strategies.sampled_from([int,
                                                                  float])}),
    os.path.join: empty.dictionaries,
    str: empty.dictionaries,
    sum: empty.dictionaries,
}
to_transparent_functions_args = transparent_functions_args.__getitem__
to_transparent_functions_kwargs = transparent_functions_kwargs.__getitem__
projectors = strategies.sampled_from([add, and_, max, min, or_,
                                      os.path.join, sub, xor])
projectors_domains = {add: [lists, numbers, strings, tuples],
                      and_: [sets],
                      max: sortable_domains,
                      min: sortable_domains,
                      or_: [sets],
                      os.path.join: [paths_names_parts],
                      sub: [numbers, sets],
                      xor: [sets]}
projectors_domains = dict(zip(projectors_domains.keys(),
                              map(strategies.sampled_from,
                                  projectors_domains.values())))
projectors_domains_initials = {
    (add, lists): empty.lists,
    (add, numbers): strategies.just(0),
    (and_, sets): sets,
    (add, strings): empty.strings,
    (add, tuples): empty.tuples,
    (max, real_numbers): strategies.just(float('-inf')),
    (max, sets): empty.sets,
    (max, strings): empty.strings,
    (min, real_numbers): strategies.just(float('inf')),
    (min, sets): empty.sets,
    (min, strings): empty.strings,
    (or_, sets): empty.sets,
    (os.path.join, paths_names_parts): empty.strings,
    (sub, numbers): strategies.just(0),
    (sub, sets): empty.sets,
    (xor, sets): empty.sets
}
to_projectors_domains = projectors_domains.__getitem__
to_projectors_domains_initials = projectors_domains_initials.__getitem__



from itertools import tee
from typing import (Any,
                    Iterable)

from lz.iterating import (last,
                          slider,
                          trailer)
from tests.utils import (are_iterables_similar,
                         capacity)


def test_capacity(non_empty_iterable: Iterable[Any],
                  natural_number: int) -> None:
    original, target = tee(non_empty_iterable)
    trail = trailer(natural_number)

    result = trail(target)

    if capacity(original) < natural_number:
        assert capacity(result) < natural_number
    else:
        assert capacity(result) == natural_number


def test_elements(non_empty_iterable: Iterable[Any],
                  natural_number: int) -> None:
    original, target = tee(non_empty_iterable)
    trail = trailer(natural_number)
    slide = slider(natural_number)

    result = trail(target)
    slides = slide(original)

    assert are_iterables_similar(result, last(slides))



from typing import (Any,
                    Iterable)

from lz import right
from lz.iterating import last


def test_basic(iterable: Iterable[Any],
               object_: Any) -> None:
    attach = right.attacher(object_)

    result = attach(iterable)

    assert last(result) is object_



from itertools import tee
from typing import (Any,
                    Iterable)

from lz.iterating import (first,
                          last,
                          reverse)


def test_basic(non_empty_iterable: Iterable[Any]) -> None:
    original, target = tee(non_empty_iterable)

    result = last(target)

    assert result is first(reverse(original))



from typing import (Any,
                    Dict,
                    Tuple)

from lz.functional import to_constant


def test_basic(object_: Any,
               positional_arguments: Tuple,
               keyword_arguments: Dict[str, Any]) -> None:
    constant = to_constant(object_)

    result = constant(*positional_arguments, **keyword_arguments)

    assert result is object_



from typing import (Callable,
                    Dict,
                    Tuple)

from lz.functional import pack
from lz.hints import (Domain,
                      Range)


def test_basic(transparent_function: Callable[..., Range],
               transparent_function_args: Tuple[Domain, ...],
               transparent_function_kwargs: Dict[str, Domain]) -> None:
    packed = pack(transparent_function)

    result = packed(transparent_function_args, transparent_function_kwargs)

    assert result == transparent_function(*transparent_function_args,
                                          **transparent_function_kwargs)



from typing import (Callable,
                    Dict,
                    Tuple)

from lz import left
from lz.hints import (Domain,
                      Range)


def test_basic(transparent_function: Callable[..., Range],
               transparent_function_args: Tuple[Domain, ...],
               transparent_function_first_args_part: Tuple[Domain, ...],
               transparent_function_second_args_part: Tuple[Domain, ...],
               transparent_function_kwargs: Dict[str, Domain],
               transparent_function_first_kwargs_part: Dict[str, Domain],
               transparent_function_second_kwargs_part: Dict[str, Domain]
               ) -> None:
    applied = left.applier(transparent_function,
                           *transparent_function_first_args_part,
                           **transparent_function_first_kwargs_part)

    result = applied(*transparent_function_second_args_part,
                     **transparent_function_second_kwargs_part)

    assert result == transparent_function(*transparent_function_args,
                                          **transparent_function_kwargs)



from itertools import tee
from typing import Iterable

import pytest

from lz.hints import (Domain,
                      Predicate)
from lz.iterating import kicker
from tests.utils import (are_iterables_similar,
                         capacity)


def test_basic(iterable: Iterable[Domain]) -> None:
    original, target = tee(iterable)
    kick = kicker()

    result = kick(target)
    result_iterator = iter(result)

    for original_element in original:
        if not original_element:
            assert original_element is next(result_iterator)
            break

    for result_element, original_element in zip(result_iterator, original):
        assert result_element is original_element

    with pytest.raises(StopIteration):
        next(result_iterator)


def test_false_predicate(iterable: Iterable[Domain],
                         false_predicate: Predicate) -> None:
    original, target = tee(iterable)
    keep_all = kicker(false_predicate)

    keep_all_result = keep_all(target)

    assert are_iterables_similar(keep_all_result, original)


def test_true_predicate(iterable: Iterable[Domain],
                        true_predicate: Predicate) -> None:
    kick_all = kicker(true_predicate)

    kick_all_result = kick_all(iterable)

    assert capacity(kick_all_result) == 0



from .definitions import (built_in_functions,
                          callables,
                          classes,
                          functions,
                          methods,
                          methods_descriptors,
                          wrappers_descriptors)
from .functional import (false_predicates,
                         maps,
                         maps_arguments,
                         maps_lists,
                         maps_lists_arguments,
                         predicates,
                         predicates_arguments,
                         projectors,
                         suitable_maps,
                         to_one_of_suitable_maps,
                         to_projectors_domains,
                         to_projectors_domains_initials,
                         to_transparent_functions_args,
                         to_transparent_functions_kwargs,
                         transparent_functions,
                         true_predicates)
from .iterating import groupers_keys
from .literals import empty
from .literals.base import (hashables,
                            indices,
                            keywords_arguments,
                            objects,
                            positionals_arguments,
                            slices,
                            sortable_iterables)
from .literals.factories import (to_integers,
                                 to_iterables)



from hypothesis import strategies

dictionaries = strategies.fixed_dictionaries({})
iterables = strategies.iterables(strategies.nothing())
lists = strategies.lists(strategies.nothing())
sets = strategies.sets(strategies.nothing())
strings = strategies.just('')
tuples = strategies.tuples()



from typing import (Any,
                    Dict,
                    Iterable,
                    Tuple)

import pytest

from tests import strategies
from tests.utils import find


@pytest.fixture(scope='function')
def object_() -> Any:
    return find(strategies.objects)


@pytest.fixture(scope='function')
def positional_arguments() -> Tuple[Any, ...]:
    return find(strategies.positionals_arguments)


@pytest.fixture(scope='function')
def keyword_arguments() -> Dict[str, Any]:
    return find(strategies.keywords_arguments)


@pytest.fixture(scope='function')
def min_iterables_size() -> int:
    return find(strategies.to_integers(0, 100))


@pytest.fixture(scope='function')
def natural_number() -> int:
    return find(strategies.to_integers(0))


@pytest.fixture(scope='function')
def iterable(min_iterables_size: int) -> Iterable[Any]:
    return find(strategies.to_iterables(strategies.hashables,
                                        min_size=min_iterables_size))


@pytest.fixture(scope='function')
def empty_iterable() -> Iterable[Any]:
    return find(strategies.empty.iterables)


@pytest.fixture(scope='function')
def non_empty_iterable() -> Iterable[Any]:
    return find(strategies.to_iterables(strategies.hashables,
                                        min_size=1))


@pytest.fixture(scope='function')
def sortable_iterable() -> Iterable[Any]:
    return find(strategies.sortable_iterables)



import random
from typing import (Callable,
                    Dict,
                    Iterable,
                    List,
                    Tuple)

import pytest
from hypothesis.searchstrategy import SearchStrategy

from lz.hints import (Domain,
                      Intermediate,
                      Map,
                      Predicate,
                      Range)
from tests import strategies
from tests.utils import find


@pytest.fixture(scope='session')
def true_predicate() -> Predicate:
    return find(strategies.true_predicates)


@pytest.fixture(scope='session')
def false_predicate() -> Predicate:
    return find(strategies.false_predicates)


@pytest.fixture(scope='session')
def predicate() -> Predicate:
    return find(strategies.predicates)


@pytest.fixture(scope='session')
def left_predicate() -> Predicate:
    return find(strategies.predicates)


@pytest.fixture(scope='session')
def mid_predicate() -> Predicate:
    return find(strategies.predicates)


@pytest.fixture(scope='session')
def right_predicate() -> Predicate:
    return find(strategies.predicates)


@pytest.fixture(scope='session')
def predicate_argument() -> Domain:
    return find(strategies.predicates_arguments)


@pytest.fixture(scope='function')
def map_() -> Map:
    return find(strategies.maps)


@pytest.fixture(scope='function')
def map_argument() -> Domain:
    return find(strategies.maps_arguments)


@pytest.fixture(scope='function')
def maps() -> List[Map]:
    return find(strategies.maps_lists)


@pytest.fixture(scope='function')
def maps_arguments() -> List[Domain]:
    return find(strategies.maps_lists_arguments)


@pytest.fixture(scope='function')
def next_map(map_: Map[Domain, Range]) -> Map[Range, Intermediate]:
    return find(strategies.to_one_of_suitable_maps(map_))


@pytest.fixture(scope='function')
def last_map(next_map: Map[Range, Intermediate]) -> Map[Intermediate, Range]:
    return find(strategies.to_one_of_suitable_maps(next_map))


@pytest.fixture(scope='function')
def suitable_maps() -> Tuple[Map, ...]:
    return find(strategies.suitable_maps)


@pytest.fixture(scope='function')
def next_suitable_map(suitable_maps: Tuple[Map, ...]) -> Map:
    return find(strategies.to_one_of_suitable_maps(suitable_maps[0]))


@pytest.fixture(scope='function')
def transparent_function() -> Callable[..., Range]:
    return find(strategies.transparent_functions)


@pytest.fixture(scope='function')
def transparent_function_args(transparent_function: Callable[..., Range]
                              ) -> Tuple[Domain, ...]:
    return find(strategies.to_transparent_functions_args(transparent_function))


@pytest.fixture(scope='function')
def transparent_function_kwargs(transparent_function: Callable[..., Range]
                                ) -> Dict[str, Domain]:
    return find(strategies
                .to_transparent_functions_kwargs(transparent_function))


@pytest.fixture(scope='function')
def transparent_function_args_count(
        transparent_function_args: Tuple[Domain, ...]) -> int:
    return len(transparent_function_args)


@pytest.fixture(scope='function')
def transparent_function_kwargs_count(
        transparent_function_kwargs: Dict[str, Domain]) -> int:
    return len(transparent_function_kwargs)


@pytest.fixture(scope='function')
def transparent_function_applied_kwargs_count(
        transparent_function_kwargs_count: int) -> int:
    return find(strategies.to_integers(
            min_value=0,
            max_value=transparent_function_kwargs_count))


@pytest.fixture(scope='function')
def transparent_function_applied_args_count(
        transparent_function_args_count: int) -> int:
    return find(strategies.to_integers(
            min_value=0,
            max_value=transparent_function_args_count))


@pytest.fixture(scope='function')
def transparent_function_first_args_part(
        transparent_function_args: Tuple[Domain, ...],
        transparent_function_applied_args_count: int) -> Tuple[Domain, ...]:
    return transparent_function_args[:transparent_function_applied_args_count]


@pytest.fixture(scope='function')
def transparent_function_first_kwargs_part(
        transparent_function_kwargs: Dict[str, Domain],
        transparent_function_applied_kwargs_count: int) -> Dict[str, Domain]:
    keys = random.sample(list(transparent_function_kwargs),
                         transparent_function_applied_kwargs_count)
    return {key: transparent_function_kwargs[key]
            for key in keys}


@pytest.fixture(scope='function')
def transparent_function_second_args_part(
        transparent_function_args: Tuple[Domain, ...],
        transparent_function_applied_args_count: int) -> Tuple[Domain, ...]:
    return transparent_function_args[transparent_function_applied_args_count:]


@pytest.fixture(scope='function')
def transparent_function_second_kwargs_part(
        transparent_function_kwargs: Dict[str, Domain],
        transparent_function_first_kwargs_part) -> len:
    return {key: value
            for key, value in transparent_function_kwargs.items()
            if key not in transparent_function_first_kwargs_part}


@pytest.fixture(scope='function')
def projector() -> Callable[[Domain, Domain], Domain]:
    return find(strategies.projectors)


@pytest.fixture(scope='function')
def projector_domain(projector: Callable[[Domain, Domain], Domain]
                     ) -> SearchStrategy:
    return find(strategies.to_projectors_domains(projector))


@pytest.fixture(scope='function')
def projector_domain_element(projector_domain: SearchStrategy) -> Domain:
    return find(projector_domain)


@pytest.fixture(scope='function')
def projector_initial(projector: Callable[[Domain, Domain], Domain],
                      projector_domain: SearchStrategy) -> Range:
    return find(strategies.to_projectors_domains_initials((projector,
                                                           projector_domain)))


@pytest.fixture(scope='function')
def projector_iterable(projector_domain: SearchStrategy) -> Iterable[Domain]:
    return find(strategies.to_iterables(projector_domain,
                                        min_size=1))



import importlib
import inspect
from types import ModuleType
from typing import (Any,
                    Union)

from hypothesis import strategies
from hypothesis.searchstrategy import SearchStrategy
from paradigm.definitions import (is_supported,
                                  stdlib_modules_names,
                                  to_contents,
                                  unsupported)
from paradigm.hints import (MethodDescriptorType,
                            WrapperDescriptorType)

stdlib_modules = list(map(importlib.import_module,
                          stdlib_modules_names
                          - unsupported.stdlib_modules_names))
modules = (strategies.sampled_from(stdlib_modules)
           .filter(is_supported))


def flatten_module_or_class(object_: Union[ModuleType, type]
                            ) -> SearchStrategy:
    return strategies.sampled_from(to_contents(object_))


modules_callables = (modules.flatmap(flatten_module_or_class)
                     .filter(callable))
classes = (modules_callables.filter(inspect.isclass)
           .filter(is_supported))
classes_callables = (classes.flatmap(flatten_module_or_class)
                     .filter(callable))
methods = classes_callables.filter(inspect.isfunction)


def is_method_descriptor(object_: Any) -> bool:
    return isinstance(object_, MethodDescriptorType)


methods_descriptors = (classes_callables.filter(is_method_descriptor)
                       .filter(is_supported))


def is_wrapper_descriptor(object_: Any) -> bool:
    return isinstance(object_, WrapperDescriptorType)


wrappers_descriptors = (classes_callables.filter(is_wrapper_descriptor)
                        .filter(is_supported))
functions = modules_callables.filter(inspect.isfunction)
built_in_functions = (modules_callables.filter(inspect.isbuiltin)
                      .filter(is_supported))
callables = (built_in_functions
             | classes
             | functions
             | methods
             | methods_descriptors
             | wrappers_descriptors)



from typing import (Any,
                    Callable,
                    Tuple)

from lz.functional import (compose,
                           identity)
from lz.hints import (Domain,
                      Intermediate,
                      Map,
                      Range)


def test_base_case(callable_: Callable[..., Any]) -> None:
    composition = compose(callable_)

    assert composition is callable_


def test_step(suitable_maps: Tuple[Map[Domain, Intermediate], ...],
              next_suitable_map: Map[Intermediate, Range],
              map_argument: Domain) -> None:
    composition = compose(*suitable_maps)
    next_composition = compose(next_suitable_map, *suitable_maps)

    result = composition(map_argument)
    next_result = next_composition(map_argument)

    assert next_result == next_suitable_map(result)


def test_identity(map_: Map[Domain, Range],
                  map_argument: Domain) -> None:
    left_composition = compose(identity, map_)
    right_composition = compose(map_, identity)

    left_composition_result = left_composition(map_argument)
    right_composition_result = right_composition(map_argument)

    assert left_composition_result == right_composition_result


def test_associativity(last_map: Map[Intermediate, Range],
                       next_map: Map[Range, Intermediate],
                       map_: Map[Domain, Range],
                       map_argument: Domain) -> None:
    left_composition = compose(compose(last_map, next_map), map_)
    right_composition = compose(last_map, compose(next_map, map_))

    left_composition_result = left_composition(map_argument)
    right_composition_result = right_composition(map_argument)

    assert left_composition_result == right_composition_result



from typing import (Any,
                    Dict,
                    Tuple)

from lz.hints import Predicate
from lz.logical import negate


def test_basic(false_predicate: Predicate,
               true_predicate: Predicate,
               positional_arguments: Tuple,
               keyword_arguments: Dict[str, Any]) -> None:
    negated_false_predicate = negate(false_predicate)
    negated_true_predicate = negate(true_predicate)

    negated_false_predicate_result = negated_false_predicate(
            *positional_arguments,
            **keyword_arguments)
    negated_true_predicate_result = negated_true_predicate(
            *positional_arguments,
            **keyword_arguments)

    assert isinstance(negated_false_predicate_result, bool)
    assert isinstance(negated_true_predicate_result, bool)
    assert negated_false_predicate_result
    assert not negated_true_predicate_result


def test_involution(false_predicate: Predicate,
                    true_predicate: Predicate,
                    positional_arguments: Tuple,
                    keyword_arguments: Dict[str, Any]) -> None:
    double_negated_false_predicate = negate(negate(false_predicate))
    double_negated_true_predicate = negate(negate(true_predicate))

    double_negated_false_predicate_result = double_negated_false_predicate(
            *positional_arguments,
            **keyword_arguments)
    double_negated_true_predicate_result = double_negated_true_predicate(
            *positional_arguments,
            **keyword_arguments)

    assert isinstance(double_negated_false_predicate_result, bool)
    assert isinstance(double_negated_true_predicate_result, bool)
    assert not double_negated_false_predicate_result
    assert double_negated_true_predicate_result



from types import (BuiltinFunctionType,
                   FunctionType,
                   MethodType)
from typing import (Any,
                    Callable)

import pytest

from paradigm.hints import (MethodDescriptorType,
                            WrapperDescriptorType)
from tests import strategies
from tests.utils import find


@pytest.fixture(scope='function')
def built_in_function() -> BuiltinFunctionType:
    return find(strategies.built_in_functions)


@pytest.fixture(scope='function')
def callable_() -> Callable[..., Any]:
    return find(strategies.callables)


@pytest.fixture(scope='function')
def class_() -> type:
    return find(strategies.classes)


@pytest.fixture(scope='function')
def function() -> FunctionType:
    return find(strategies.functions)


@pytest.fixture(scope='function')
def method() -> MethodType:
    return find(strategies.methods)


@pytest.fixture(scope='function')
def method_descriptor() -> MethodDescriptorType:
    return find(strategies.methods_descriptors)


@pytest.fixture(scope='function')
def wrapper_descriptor() -> WrapperDescriptorType:
    return find(strategies.methods_descriptors)



from typing import Any

import pytest

from lz.iterating import expand


def test_basic(object_: Any) -> None:
    result = expand(object_)
    result_iterator = iter(result)

    assert next(result_iterator) is object_

    with pytest.raises(StopIteration):
        next(result_iterator)



from typing import List

from lz.functional import combine
from lz.hints import (Domain,
                      Map)
from tests.utils import capacity


def test_capacity(maps: List[Map],
                  maps_arguments: List[Domain]) -> None:
    combined = combine(maps)

    result = combined(maps_arguments)

    assert capacity(result) == min(len(maps), len(maps_arguments))


def test_elements(maps: List[Map],
                  maps_arguments: List[Domain]) -> None:
    combined = combine(maps)

    result = combined(maps_arguments)

    assert all(element == map_(map_argument)
               for element, map_, map_argument in zip(result,
                                                      maps, maps_arguments))



from itertools import (islice,
                       tee)
from typing import (Any,
                    Iterable)

import pytest

from lz.iterating import slider


def test_basic(iterable: Iterable[Any],
               slider_size: int) -> None:
    original, target = tee(iterable)
    slide = slider(slider_size)

    result = slide(target)
    result_iterator = iter(result)

    original_elements = tuple(islice(original, slider_size))
    assert original_elements == next(result_iterator)
    for result_element, original_element in zip(result_iterator, original):
        original_elements = original_elements[1:] + (original_element,)
        assert result_element == original_elements

    with pytest.raises(StopIteration):
        next(original)

    with pytest.raises(StopIteration):
        next(result_iterator)



from lz.hints import (Domain,
                      Predicate)
from lz.logical import exclusive_disjoin


def test_idempotency(predicate: Predicate,
                     predicate_argument: Domain) -> None:
    self_disjunction = exclusive_disjoin(predicate)

    result = self_disjunction(predicate_argument)

    assert result is predicate(predicate_argument)


def test_neutral_element(predicate: Predicate,
                         false_predicate: Predicate,
                         predicate_argument: Domain) -> None:
    left_disjunction = exclusive_disjoin(predicate, false_predicate)
    right_disjunction = exclusive_disjoin(false_predicate, predicate)

    left_result = left_disjunction(predicate_argument)
    right_result = right_disjunction(predicate_argument)

    assert left_result is right_result is predicate(predicate_argument)


def test_commutativity(left_predicate: Predicate,
                       right_predicate: Predicate,
                       predicate_argument: Domain) -> None:
    left_disjunction = exclusive_disjoin(left_predicate, right_predicate)
    right_disjunction = exclusive_disjoin(right_predicate, left_predicate)

    left_result = left_disjunction(predicate_argument)
    right_result = right_disjunction(predicate_argument)

    assert left_result is right_result


def test_associativity(left_predicate: Predicate,
                       mid_predicate: Predicate,
                       right_predicate: Predicate,
                       predicate_argument: Domain) -> None:
    left_disjunction = exclusive_disjoin(left_predicate, right_predicate)
    right_disjunction = exclusive_disjoin(right_predicate, left_predicate)

    left_result = left_disjunction(predicate_argument)
    right_result = right_disjunction(predicate_argument)

    assert left_result is right_result



from typing import (Any,
                    Callable,
                    Dict)

from lz.functional import flip
from lz.hints import Domain


def test_basic(transparent_function: Callable[..., Any],
               transparent_function_args: Domain,
               transparent_function_kwargs: Dict[str, Domain]) -> None:
    flipped = flip(transparent_function)

    original_result = transparent_function(*transparent_function_args,
                                           **transparent_function_kwargs)
    flipped_result = flipped(*reversed(transparent_function_args),
                             **transparent_function_kwargs)

    assert flipped_result == original_result


def test_involution(transparent_function: Callable[..., Any],
                    transparent_function_args: Domain,
                    transparent_function_kwargs: Dict[str, Domain]) -> None:
    double_flipped = flip(flip(transparent_function))

    original_result = transparent_function(*transparent_function_args,
                                           **transparent_function_kwargs)
    double_flipped_result = double_flipped(*transparent_function_args,
                                           **transparent_function_kwargs)

    assert double_flipped_result == original_result



from itertools import tee
from typing import (Callable,
                    Iterable)

from lz import (left,
                right)
from lz.hints import (Domain,
                      Range)


def test_base_case(projector: Callable[[Domain, Range], Range],
                   projector_initial: Range,
                   empty_iterable: Iterable[Domain]) -> None:
    fold = right.folder(projector, projector_initial)

    result = fold(empty_iterable)

    assert result is projector_initial


def test_step(projector: Callable[[Domain, Range], Range],
              projector_initial: Range,
              projector_iterable: Iterable[Domain],
              projector_domain_element: Domain) -> None:
    first_target, second_target = tee(projector_iterable)
    fold = right.folder(projector, projector_initial)
    attach = left.attacher(projector_domain_element)

    result = fold(first_target)
    next_result = fold(attach(second_target))

    assert next_result == projector(projector_domain_element, result)



from itertools import tee
from typing import Iterable

import pytest

from lz.hints import (Domain,
                      Predicate)
from lz.iterating import grabber
from tests.utils import (are_iterables_similar,
                         capacity)


def test_basic(iterable: Iterable[Domain]) -> None:
    original, target = tee(iterable)
    grab = grabber()

    result = grab(target)
    result_iterator = iter(result)

    for result_element, original_element in zip(result_iterator, original):
        if not original_element:
            break
        assert result_element is original_element

    with pytest.raises(StopIteration):
        next(result_iterator)


def test_false_predicate(iterable: Iterable[Domain],
                         false_predicate: Predicate) -> None:
    skip_all = grabber(false_predicate)

    skip_all_result = skip_all(iterable)

    assert capacity(skip_all_result) == 0


def test_true_predicate(iterable: Iterable[Domain],
                        true_predicate: Predicate) -> None:
    original, target = tee(iterable)
    grab_all = grabber(true_predicate)

    grab_all_result = grab_all(target)

    assert are_iterables_similar(grab_all_result, original)



import functools
import itertools
from typing import (Callable,
                    Iterable)

from . import left
from .functional import (compose,
                         flip)
from .hints import (Domain,
                    Map,
                    Range)
from .iterating import (expand,
                        reverse)


def accumulator(function: Callable[[Domain, Range], Range],
                initial: Range
                ) -> Map[Iterable[Domain], Iterable[Iterable[Range]]]:
    """
    Returns function that yields cumulative results of given binary function
    starting from given initial object in direction from right to left.
    """
    left_accumulator = left.accumulator(flip(function), initial)
    return compose(left_accumulator, reverse)


def attacher(object_: Domain) -> Map[Iterable[Domain], Iterable[Domain]]:
    """
    Returns function that appends given object to iterable.
    """

    def attach(iterable: Iterable[Domain]) -> Iterable[Domain]:
        yield from itertools.chain(iterable, expand(object_))

    return attach


def folder(function: Callable[[Domain, Range], Range],
           initial: Range) -> Map[Iterable[Domain], Range]:
    """
    Returns function that cumulatively applies given binary function
    starting from given initial object in direction from right to left.
    """
    left_folder = left.folder(flip(function), initial)
    return compose(left_folder, reverse)


def applier(function: Callable[..., Range],
            *args: Domain,
            **kwargs: Domain) -> Callable[..., Range]:
    """
    Returns function that behaves like given function
    with given arguments partially applied.
    Positional arguments will be applied from the right end.
    """
    args = args[::-1]

    @functools.wraps(function)
    def applied(*rest_args, **rest_kwargs) -> Range:
        return function(*rest_args, *args, **kwargs, **rest_kwargs)

    return applied



from itertools import tee
from typing import (Any,
                    Hashable,
                    Iterable)

from lz.hints import Map
from lz.iterating import grouper


def test_basic(iterable: Iterable[Any],
               grouper_key: Map[Any, Hashable]) -> None:
    original, target = tee(iterable)

    result = grouper(grouper_key)(target)
    result_list = list(result)

    assert all(element in group
               for element in original
               for key, group in result_list
               if grouper_key(element) == key)



from typing import (Callable,
                    TypeVar)

from typing_extensions import Protocol

Domain = TypeVar('Domain')
Intermediate = TypeVar('Intermediate')
Range = TypeVar('Range')
Map = Callable[[Domain], Range]
Operator = Map[Domain, Domain]
Predicate = Map[Domain, bool]


class Sortable(Protocol):
    def __lt__(self, other: 'Sortable') -> bool:
        pass



import string

from hypothesis import strategies
from hypothesis.searchstrategy import SearchStrategy

from tests.configs import MAX_ITERABLES_SIZE
from .factories import (to_dictionaries,
                        to_frozensets,
                        to_iterables,
                        to_lists,
                        to_sets,
                        to_strings)

strings = to_strings()
integers = (strategies.booleans()
            | strategies.integers())
real_numbers = (integers
                | strategies.floats(allow_nan=False,
                                    allow_infinity=False))
numbers = (real_numbers
           | strategies.complex_numbers(allow_nan=False,
                                        allow_infinity=False))
scalars = (strategies.none()
           | numbers
           | strings)
hashables = (scalars
             | to_frozensets(strategies.deferred(lambda: hashables))
             | to_lists(strategies.deferred(lambda: hashables))
             .map(tuple))
deferred_objects = strategies.deferred(lambda: objects)
lists = to_lists(deferred_objects)
tuples = lists.map(tuple)
indices = strategies.integers(0, MAX_ITERABLES_SIZE)
slices_fields = strategies.none() | indices
slices = strategies.builds(slice,
                           slices_fields,
                           slices_fields,
                           slices_fields)
sets = to_sets(hashables)
objects = (hashables
           | slices
           | to_dictionaries(hashables, deferred_objects)
           | to_iterables(deferred_objects)
           | lists
           | sets
           | tuples)


def extend_json(children: SearchStrategy) -> SearchStrategy:
    return (strategies.lists(children)
            | to_dictionaries(to_strings(string.printable),
                              children))


json_serializable_objects = strategies.recursive(
        strategies.none()
        | real_numbers
        | to_strings(string.printable),
        extend_json)
positionals_arguments = tuples
keywords_arguments = to_dictionaries(strings, objects)

sortable_domains = [real_numbers, sets, strings]
sortable_iterables = strategies.one_of(*map(to_iterables, sortable_domains))



from itertools import tee
from typing import (Any,
                    Iterable)

from lz.iterating import first


def test_basic(non_empty_iterable: Iterable[Any]) -> None:
    original, target = tee(non_empty_iterable)

    result = first(target)

    assert result is next(original)



from itertools import tee
from typing import (Callable,
                    Iterable)

from lz import left
from lz.hints import (Domain,
                      Range)
from lz.iterating import (first,
                          last)


def test_first(projector: Callable[[Range, Domain], Range],
               projector_initial: Range,
               projector_iterable: Iterable[Domain]) -> None:
    accumulate = left.accumulator(projector, projector_initial)

    result = accumulate(projector_iterable)

    assert first(result) is projector_initial


def test_last(projector: Callable[[Range, Domain], Range],
              projector_initial: Range,
              projector_iterable: Iterable[Domain]) -> None:
    first_target, second_target = tee(projector_iterable)
    accumulate = left.accumulator(projector, projector_initial)
    fold = left.folder(projector, projector_initial)

    result = accumulate(first_target)
    fold_result = fold(second_target)

    assert last(result) == fold_result



from typing import (Any,
                    Hashable)

import pytest

from lz.hints import Map
from tests import strategies
from tests.utils import find


@pytest.fixture(scope='function')
def grouper_key() -> Map[Any, Hashable]:
    return find(strategies.groupers_keys)


@pytest.fixture(scope='function')
def cutter_slice() -> slice:
    return find(strategies.slices)


@pytest.fixture(scope='function')
def chopper_size() -> int:
    return find(strategies.indices)


@pytest.fixture(scope='function')
def slider_size(min_iterables_size: int) -> int:
    return find(strategies.to_integers(0, min_iterables_size))



from itertools import (islice,
                       tee)
from typing import (Any,
                    Iterable)

from lz.iterating import cutter
from tests.configs import MAX_ITERABLES_SIZE
from tests.utils import (are_iterables_similar,
                         capacity)


def test_capacity(iterable: Iterable[Any],
                  cutter_slice: slice) -> None:
    cut = cutter(cutter_slice)
    result = cut(iterable)

    assert capacity(result) <= slice_to_size(cutter_slice)


def test_elements(iterable: Iterable[Any],
                  cutter_slice: slice) -> None:
    original, target = tee(iterable)

    cut = cutter(cutter_slice)
    result = cut(target)

    assert are_iterables_similar(result,
                                 islice(original,
                                        cutter_slice.start,
                                        cutter_slice.stop,
                                        cutter_slice.step))


def slice_to_size(slice_: slice) -> int:
    start, stop, step = slice_.indices(MAX_ITERABLES_SIZE)
    return max(1 + (stop - 1 - start) // step, 0)



from operator import (not_,
                      xor)

from . import left
from .functional import (cleave,
                         compose)
from .hints import Predicate


def conjoin(*predicates: Predicate) -> Predicate:
    """
    Returns conjunction of given predicates.
    """
    return compose(all, cleave(predicates))


def disjoin(*predicates: Predicate) -> Predicate:
    """
    Returns disjunction of given predicates.
    """
    return compose(any, cleave(predicates))


def exclusive_disjoin(*predicates: Predicate) -> Predicate:
    """
    Returns exclusive disjunction of given predicates.
    """
    return compose(left.folder(xor, False), cleave(predicates))


def negate(predicate: Predicate) -> Predicate:
    """
    Returns negated version of given predicate.
    """
    return compose(not_, predicate)



from functools import partial

from hypothesis import strategies

from tests.configs import MAX_ITERABLES_SIZE

to_integers = strategies.integers
limit_max_size = partial(partial,
                         max_size=MAX_ITERABLES_SIZE)
to_dictionaries = limit_max_size(strategies.dictionaries)
to_frozensets = limit_max_size(strategies.frozensets)
to_iterables = limit_max_size(strategies.iterables)
to_lists = limit_max_size(strategies.lists)
to_sets = limit_max_size(strategies.sets)
to_strings = limit_max_size(strategies.text)



from itertools import tee
from typing import (Any,
                    Iterable)

from lz.iterating import reverse
from tests.utils import are_iterables_similar


def test_basic(iterable: Iterable[Any]) -> None:
    original, target = tee(iterable)

    result = reverse(target)

    assert are_iterables_similar(result, list(original)[::-1])


def test_involution(iterable: Iterable[Any]) -> None:
    original, target = tee(iterable)

    result = reverse(reverse(target))

    assert are_iterables_similar(result, original)



from itertools import tee
from typing import Iterable

from lz.hints import Domain
from lz.iterating import (slider,
                          sorter)
from tests.utils import (are_iterables_similar,
                         capacity)


def test_order(sortable_iterable: Iterable[Domain]) -> None:
    sort = sorter()

    result = sort(sortable_iterable)

    elements_pairs = iter(slider(2)(result))
    try:
        element, next_element = next(elements_pairs)
    except ValueError:
        # iterable with no elements
        # or with single element are considered ordered
        return
    else:
        assert objects_are_partially_ordered(element, next_element)
    assert all(objects_are_partially_ordered(element, next_element)
               for element, next_element in elements_pairs)


def objects_are_partially_ordered(left_object: Domain,
                                  right_object: Domain) -> bool:
    return left_object < right_object or not (right_object < left_object)


def test_capacity(sortable_iterable: Iterable[Domain]) -> None:
    original, target = tee(sortable_iterable)
    sort = sorter()

    result = sort(target)

    assert capacity(result) == capacity(original)


def test_elements(sortable_iterable: Iterable[Domain]) -> None:
    original, target = tee(sortable_iterable)
    sort = sorter()

    result = sort(target)
    original_list = list(original)

    assert all(element in original_list
               for element in result)


def test_idempotency(sortable_iterable: Iterable[Domain]) -> None:
    first_target, second_target = tee(sortable_iterable)
    sort = sorter()

    sorted_result = sort(first_target)
    double_sorted_result = sort(sort(second_target))

    assert are_iterables_similar(double_sorted_result, sorted_result)



import functools
import inspect
import itertools
from types import MappingProxyType
from typing import (Any,
                    Callable,
                    Dict,
                    Iterable,
                    Mapping,
                    Tuple,
                    Union)

from paradigm import signatures

from .hints import (Domain,
                    Intermediate,
                    Map,
                    Range)


def identity(argument: Domain) -> Domain:
    """
    Returns object itself.
    """
    return argument


def compose(last_function: Map[Any, Range],
            *front_functions: Callable[..., Any]) -> Callable[..., Range]:
    """
    Returns functions composition.
    """

    def binary_compose(left_function: Map[Intermediate, Range],
                       right_function: Callable[..., Intermediate]
                       ) -> Callable[..., Range]:
        def composition(*args, **kwargs):
            return left_function(right_function(*args, **kwargs))

        return composition

    functions = (last_function,) + front_functions
    return functools.reduce(binary_compose, functions)


def combine(maps: Iterable[Map]) -> Map[Iterable[Domain], Iterable[Range]]:
    """
    Returns function that applies each map to corresponding argument.
    """

    def combined(arguments: Iterable[Domain]) -> Iterable[Range]:
        yield from (map_(argument)
                    for map_, argument in zip(maps, arguments))

    return combined


class Curry:
    def __init__(self,
                 callable_: Callable[..., Range],
                 signature: signatures.Base,
                 *args: Domain,
                 **kwargs: Domain) -> None:
        self.callable_ = callable_
        self.signature = signature
        self.args = args
        self.kwargs = kwargs

    def __call__(self, *args: Domain, **kwargs: Domain
                 ) -> Union['Curry', Range]:
        args = self.args + args
        kwargs = {**self.kwargs, **kwargs}
        try:
            return self.callable_(*args, **kwargs)
        except TypeError:
            if not self.signature.has_unset_parameters(*args, **kwargs):
                raise
            return type(self)(self.callable_, self.signature, *args, **kwargs)

    def __eq__(self, other: 'Curry') -> bool:
        if not isinstance(other, Curry):
            return NotImplemented
        return (self.callable_ is other.callable_
                and self.signature == other.signature
                and self.args == other.args
                and self.kwargs == other.kwargs)

    def __repr__(self) -> str:
        result = ('<curried {callable_}'
                  .format(callable_=self.callable_))
        arguments_strings = list(
                itertools.chain(map(str, self.args),
                                itertools.starmap('{}={}'.format,
                                                  self.kwargs.items())))
        if arguments_strings:
            result += (' with partially applied {arguments}'
                       .format(arguments=', '.join(arguments_strings)))
        result += '>'
        return result


def unwrap(function: Callable[..., Range]) -> Tuple[Callable[..., Range],
                                                    Tuple[Domain, ...],
                                                    Dict[str, Domain]]:
    original_function = function
    try:
        function, args, kwargs = (inspect.unwrap(function.func),
                                  function.args,
                                  function.keywords)
    except AttributeError:
        return inspect.unwrap(original_function), (), {}
    else:
        if (not isinstance(function, Callable)
                or not isinstance(args, Iterable)
                or not isinstance(kwargs, Mapping)):
            return inspect.unwrap(original_function), (), {}
        return function, args, kwargs


def curry(callable_: Callable[..., Range]) -> Curry:
    """
    Returns curried version of given callable.
    """
    callable_, args, kwargs = unwrap(callable_)
    signature = signatures.factory(callable_)
    return Curry(callable_, signature, *args, **kwargs)


def pack(function: Callable[..., Range]) -> Map[Iterable[Domain], Range]:
    """
    Returns function that works with single iterable parameter
    by unpacking elements to given function.
    """

    @functools.wraps(function)
    def packed(args: Iterable[Domain],
               kwargs: Dict[str, Any] = MappingProxyType({})) -> Range:
        return function(*args, **kwargs)

    return packed


def to_constant(object_: Domain) -> Callable[..., Domain]:
    """
    Returns function that returns given object.
    """

    def constant(*_: Domain, **__: Domain) -> Domain:
        return object_

    return constant


@functools.singledispatch
def flip(object_: Callable[..., Range]) -> Callable[..., Range]:
    """
    Returns function with positional arguments flipped.
    """

    @functools.wraps(object_)
    def flipped(*args, **kwargs) -> Range:
        return object_(*reversed(args), **kwargs)

    return flipped


@flip.register(functools.partial)
def flip_partial(object_: functools.partial) -> Callable[..., Range]:
    return functools.partial(flip(object_.func),
                             *object_.args,
                             **object_.keywords)


def cleave(functions: Iterable[Callable[..., Range]]
           ) -> Callable[..., Iterable[Range]]:
    """
    Returns function that separately applies
    given functions to the same arguments.
    """

    def cleft(*args, **kwargs) -> Range:
        yield from (function(*args, **kwargs)
                    for function in functions)

    return cleft



import functools
import itertools
from typing import (Callable,
                    Iterable)

from .hints import (Domain,
                    Map,
                    Range)
from .iterating import expand


def accumulator(function: Callable[[Range, Domain], Range],
                initial: Range
                ) -> Map[Iterable[Domain], Iterable[Range]]:
    """
    Returns function that yields cumulative results of given binary function
    starting from given initial object in direction from left to right.
    """
    attach_initial = attacher(initial)

    def accumulate(iterable: Iterable[Domain]) -> Iterable[Range]:
        yield from itertools.accumulate(attach_initial(iterable), function)

    return accumulate


def attacher(object_: Domain) -> Map[Iterable[Domain], Iterable[Domain]]:
    """
    Returns function that prepends given object to iterable.
    """

    def attach(iterable: Iterable[Domain]) -> Iterable[Domain]:
        yield from itertools.chain(expand(object_), iterable)

    return attach


def folder(function: Callable[[Range, Domain], Range],
           initial: Range) -> Map[Iterable[Domain], Range]:
    """
    Returns function that cumulatively applies given binary function
    starting from given initial object in direction from left to right.
    """

    def fold(iterable: Iterable[Domain]) -> Range:
        return functools.reduce(function, iterable, initial)

    return fold


@functools.singledispatch
def applier(function: Callable[..., Range],
            *args: Domain,
            **kwargs: Domain) -> Callable[..., Range]:
    """
    Returns function that behaves like given function
    with given arguments partially applied.
    Positional arguments will be applied from the left end.
    """
    return functools.partial(function, *args, **kwargs)



from typing import Hashable

from hypothesis import strategies

from lz.functional import identity
from lz.hints import Map


def to_grouper_key(odd_order: int) -> Map[Hashable, Hashable]:
    def grouper_key(object_: Hashable) -> Hashable:
        return hash(object_) ** odd_order

    return grouper_key


def is_odd(number: int) -> bool:
    return bool(number & 1)


groupers_keys = (strategies.integers(1, 1000)
                 .filter(is_odd)
                 .map(to_grouper_key)
                 | strategies.just(identity))



from pathlib import Path

from setuptools import (find_packages,
                        setup)

import lz

project_base_url = 'https://github.com/lycantropos/lz/'

install_requires = [
    'paradigm>=0.0.3',
    'typing_extensions>=3.6.5',
]
setup_requires = [
    'pytest-runner>=4.2'
]
tests_require = [
    'pytest>=3.8.1',
    'pytest-cov>=2.6.0',
    'hypothesis>=3.73.1',
]

setup(name='lz',
      packages=find_packages(exclude=('tests',)),
      version=lz.__version__,
      description=lz.__doc__,
      long_description=Path('README.md').read_text(),
      long_description_content_type='text/markdown',
      author='Azat Ibrakov',
      author_email='azatibrakov@gmail.com',
      url=project_base_url,
      download_url=project_base_url + 'archive/master.zip',
      python_requires='>=3.5.3',
      install_requires=install_requires,
      setup_requires=setup_requires,
      tests_require=tests_require)



from types import (BuiltinFunctionType,
                   FunctionType,
                   MethodType)
from typing import (Any,
                    Callable)

from paradigm.hints import (MethodDescriptorType,
                            WrapperDescriptorType)

from lz.functional import (Curry,
                           curry)


def test_basic(built_in_function: BuiltinFunctionType,
               class_: type,
               function: FunctionType,
               method: MethodType,
               method_descriptor: MethodDescriptorType,
               wrapper_descriptor: WrapperDescriptorType) -> None:
    built_in_function_result = curry(built_in_function)
    class_result = curry(class_)
    function_result = curry(function)
    method_result = curry(method)
    method_descriptor_result = curry(method_descriptor)
    wrapper_descriptor_result = curry(wrapper_descriptor)

    assert isinstance(built_in_function_result, Curry)
    assert isinstance(class_result, Curry)
    assert isinstance(function_result, Curry)
    assert isinstance(method_result, Curry)
    assert isinstance(method_descriptor_result, Curry)
    assert isinstance(wrapper_descriptor_result, Curry)


def test_call(callable_: Callable[..., Any]) -> None:
    result = curry(callable_)

    assert (not result.signature.has_unset_parameters()
            or result() == result)



from lz.hints import (Domain,
                      Predicate)
from lz.logical import conjoin


def test_idempotency(predicate: Predicate,
                     predicate_argument: Domain) -> None:
    self_conjunction = conjoin(predicate)

    result = self_conjunction(predicate_argument)

    assert result is predicate(predicate_argument)


def test_absorbing_element(predicate: Predicate,
                           false_predicate: Predicate,
                           predicate_argument: Domain) -> None:
    left_conjunction = conjoin(predicate, false_predicate)
    right_conjunction = conjoin(false_predicate, predicate)

    left_result = left_conjunction(predicate_argument)
    right_result = right_conjunction(predicate_argument)

    assert left_result is right_result is false_predicate(predicate_argument)


def test_neutral_element(predicate: Predicate,
                         true_predicate: Predicate,
                         predicate_argument: Domain) -> None:
    left_conjunction = conjoin(predicate, true_predicate)
    right_conjunction = conjoin(true_predicate, predicate)

    left_result = left_conjunction(predicate_argument)
    right_result = right_conjunction(predicate_argument)

    assert left_result is right_result is predicate(predicate_argument)


def test_commutativity(left_predicate: Predicate,
                       right_predicate: Predicate,
                       predicate_argument: Domain) -> None:
    left_conjunction = conjoin(left_predicate, right_predicate)
    right_conjunction = conjoin(right_predicate, left_predicate)

    left_result = left_conjunction(predicate_argument)
    right_result = right_conjunction(predicate_argument)

    assert left_result is right_result


def test_associativity(left_predicate: Predicate,
                       mid_predicate: Predicate,
                       right_predicate: Predicate,
                       predicate_argument: Domain) -> None:
    left_conjunction = conjoin(conjoin(left_predicate, right_predicate),
                               mid_predicate)
    right_conjunction = conjoin(left_predicate,
                                conjoin(mid_predicate, right_predicate))

    left_result = left_conjunction(predicate_argument)
    right_result = right_conjunction(predicate_argument)

    assert left_result is right_result



import functools
import itertools
import sys
from collections import (abc,
                         defaultdict,
                         deque)
from typing import (Hashable,
                    Iterable,
                    List,
                    Mapping,
                    Tuple)

from .functional import (cleave,
                         combine,
                         compose)
from .hints import (Domain,
                    Map,
                    Operator,
                    Predicate,
                    Range,
                    Sortable)


def mapper(map_: Map) -> Map[Iterable[Domain], Iterable[Range]]:
    """
    Returns function that applies given map to the each element of iterable.
    """
    return functools.partial(map, map_)


def sifter(predicate: Predicate = None) -> Operator[Iterable[Domain]]:
    """
    Returns function that selects elements from iterable
    which satisfy given predicate.

    If predicate is not specified than true-like objects are selected.
    """
    return functools.partial(filter, predicate)


def scavenger(predicate: Predicate = None) -> Operator[Iterable[Domain]]:
    """
    Returns function that selects elements from iterable
    which dissatisfy given predicate.

    If predicate is not specified than false-like objects are selected.
    """
    return functools.partial(itertools.filterfalse, predicate)


def separator(predicate: Predicate = None
              ) -> Map[Iterable[Domain],
                       Tuple[Iterable[Domain], Iterable[Domain]]]:
    """
    Returns function that returns pair of iterables
    first of which consists of elements that dissatisfy given predicate
    and second one consists of elements that satisfy given predicate.
    """
    return compose(tuple,
                   combine([scavenger(predicate), sifter(predicate)]),
                   copier(2))


def grabber(predicate: Predicate = None) -> Operator[Iterable[Domain]]:
    """
    Returns function that selects elements from the beginning of iterable
    while given predicate is satisfied.

    If predicate is not specified than true-like objects are selected.
    """
    if predicate is None:
        predicate = bool
    return functools.partial(itertools.takewhile, predicate)


def kicker(predicate: Predicate = None) -> Operator[Iterable[Domain]]:
    """
    Returns function that skips elements from the beginning of iterable
    while given predicate is satisfied.

    If predicate is not specified than true-like objects are skipped.
    """
    if predicate is None:
        predicate = bool
    return functools.partial(itertools.dropwhile, predicate)


def cutter(slice_: slice) -> Operator[Iterable[Domain]]:
    """
    Returns function that selects elements from iterable based on given slice.

    Slice supposed to have non-negative fields
    since it is hard to evaluate negative indices for arbitrary iterable.
    """
    start = slice_.start
    stop = slice_.stop
    step = slice_.step

    def cut(iterable: Iterable[Domain]) -> Iterable[Domain]:
        yield from itertools.islice(iterable, start, stop, step)

    return cut


def chopper(size: int) -> Map[Iterable[Domain], Iterable[Tuple[Domain, ...]]]:
    """
    Returns function that splits iterable into chunks of given size.
    """
    cut = compose(tuple, cutter(slice(size)))
    return compose(grabber(),
                   cleave(itertools.repeat(cut)),
                   iter)


def slider(size: int) -> Map[Iterable[Domain], Iterable[Tuple[Domain, ...]]]:
    """
    Returns function that slides over iterable with window of given size.
    """

    def slide(iterable: Iterable[Domain]) -> Iterable[Tuple[Domain, ...]]:
        iterator = iter(iterable)
        result = tuple(itertools.islice(iterator, size))

        def shift(previous: Tuple[Domain, ...],
                  element: Domain) -> Tuple[Domain, ...]:
            return previous[1:] + (element,)

        yield from itertools.accumulate(itertools.chain([result], iterator),
                                        shift)

    return slide


def sorter(key: Map[Domain, Sortable] = None) -> Operator[Iterable[Domain]]:
    """
    Returns function that generates sorted iterable by given key.
    """

    def sort(iterable: Iterable[Domain]) -> Iterable[Domain]:
        yield from sorted(iterable,
                          key=key)

    return sort


Group = Tuple[Hashable, Iterable[Domain]]


def grouper(key: Map[Domain, Hashable]
            ) -> Map[Iterable[Domain], Iterable[Group]]:
    """
    Returns function that groups iterable elements based on given key.
    """

    def group_by(iterable: Iterable[Domain]) -> Iterable[Group]:
        groups = defaultdict(list)  # type: Mapping[Hashable, List[Domain]]
        for element in iterable:
            groups[key(element)].append(element)
        yield from groups.items()

    return group_by


@functools.singledispatch
def reverse(iterable: Iterable[Domain]) -> Iterable[Domain]:
    """
    Returns iterable with reversed elements order.
    """
    yield from reversed(list(iterable))


if sys.version_info >= (3, 6):
    from typing import Reversible


    @reverse.register(abc.Reversible)
    def reverse_reversible(iterable: Reversible[Domain]) -> Iterable[Domain]:
        yield from reversed(iterable)


def expand(object_: Domain) -> Iterable[Domain]:
    """
    Wraps object into iterable.
    """
    yield object_


def flatten(iterable: Iterable[Iterable[Domain]]) -> Iterable[Domain]:
    """
    Returns plain iterable from iterable of iterables.
    """
    yield from itertools.chain.from_iterable(iterable)


def flatmapper(map_: Map[Domain, Iterable[Range]]
               ) -> Map[Iterable[Domain], Iterable[Range]]:
    """
    Returns function that applies map to the each element of iterable
    and flattens results.
    """
    return compose(flatten, mapper(map_))


def copier(count: int) -> Map[Iterable[Domain], Iterable[Iterable[Domain]]]:
    """
    Returns function that creates independent copies of iterable.
    """
    min_count = 0
    if count < min_count:
        raise ValueError('Count should be '
                         'not less than {min_count}, '
                         'but found {actual_count}.'
                         .format(min_count=min_count,
                                 actual_count=count))

    def copy(iterable: Iterable[Domain]) -> Iterable[Iterable[Domain]]:
        yield from itertools.tee(iterable, count)

    return copy


first = compose(next, iter)
first.__doc__ = 'Returns first element of iterable.'


def trailer(size: int) -> Operator[Iterable[Domain]]:
    """
    Returns function that selects elements from the end of iterable.
    Resulted iterable will have size not greater than given one.
    """
    return compose(iter,
                   functools.partial(deque,
                                     maxlen=size))


last = compose(first, trailer(1))
last.__doc__ = 'Returns last element of iterable.'



from itertools import tee
from typing import (Any,
                    Iterable)

from lz.iterating import (chopper,
                          flatten)
from tests.utils import are_iterables_similar


def test_size(iterable: Iterable[Any],
              chopper_size: int) -> None:
    chop = chopper(chopper_size)

    result = chop(iterable)

    assert all(len(element) <= chopper_size
               for element in result)


def test_elements(iterable: Iterable[Any],
                  chopper_size: int) -> None:
    original, target = tee(iterable)
    chop = chopper(chopper_size)

    result = chop(target)

    assert are_iterables_similar(flatten(result), original)



from typing import List

from lz.functional import cleave
from lz.hints import (Domain,
                      Map)
from tests.utils import capacity


def test_capacity(maps: List[Map],
                  map_argument: Domain) -> None:
    cleft = cleave(maps)

    result = cleft(map_argument)

    assert capacity(result) == len(maps)


def test_elements(maps: List[Map],
                  map_argument: Domain) -> None:
    cleft = cleave(maps)

    result = cleft(map_argument)

    assert all(element == map_(map_argument)
               for element, map_ in zip(result, maps))
